import time
from typing import Any, Dict, List, Optional

import numpy as np
from badger import environment
from epics import caget_many, caput, caget

from pydantic import ConfigDict, model_validator

from AWAControl import MAGNET_CONFIG
from AWAControl.diagnostics.screen import load_screen, AWAImageDiagnostic, check_same_camera_type

# get ranges and pvs for control names
ranges = {name: val["RANGE"] for name, val in MAGNET_CONFIG.items()}
pvs = {name: val["PV"] for name, val in MAGNET_CONFIG.items()}

screen_observables = [
    "Sx",
    "Sy",
    "Cx",
    "Cy",
    "total_size",
    "bb_penalty",
    "log10_total_intensity",
    "total_size",
    "center_dist_x",
    "center_dist_y",
    "center_dist",
    "mean_charge"
]

class Environment(environment.Environment):
    name = "awa"
    variables = ranges

    observables = [
        "K1_phase_readback",
        "ICT1wf",
        "ICT1",
        "FCT1wf",
        "FCT1",
        "GunAll",
        "LaserPhase"
    ] + screen_observables
    screen_name: str = None
    screen_name2: str = None # added for virtual Cathode
    image_diagnostic: Any = None
    image_diagnostic2: Any = None # added for virtual Cathode
    setpoints: Optional[dict] = None
    n_shots: int = 1

    model_config = ConfigDict(validate_assignment=True)

    def set_screen(self, screen_name=None):
        screen_name = screen_name or self.screen_name
        if self.image_diagnostic is None:
            self.image_diagnostic = load_screen(screen_name)
            
        else:
            if isinstance(self.image_diagnostic, AWAImageDiagnostic):
                if self.image_diagnostic.alias != screen_name:
                    self.image_diagnostic = load_screen(screen_name)
                    
            elif isinstance(self.image_diagnostic, dict):
                self.image_diagnostic = load_screen(screen_name)
                
            else:
                raise ValueError("image diagnostic type not handled")
                
        check_same_camera_type(self.image_diagnostic,self.image_diagnostic2)
        self.screen_name = screen_name

    def set_screen2(self, screen_name=None):
        screen_name2 = screen_name or self.screen_name2
        if self.image_diagnostic2 is None:
            self.image_diagnostic2 = load_screen(screen_name2)
            
        else:
            if isinstance(self.image_diagnostic2, AWAImageDiagnostic):
                if self.image_diagnostic2.alias != screen_name2:
                    self.image_diagnostic2 = load_screen(screen_name2)
                    
            elif isinstance(self.image_diagnostic, dict):
                self.image_diagnostic2 = load_screen(screen_name2)
                
            else:
                raise ValueError("image diagnostic type not handled")
        
        check_same_camera_type(self.image_diagnostic,self.image_diagnostic2)
        self.screen_name2 = screen_name2

        

    def get_observables(self, observable_names: List[str],**kwargs) -> Dict:
        # TODO: add in a function to archive machine state / grab LLRF amps/phases etc. that doesn't participate in control # SH: utils.py->get_GunPVs(),get_DSs()
        # TODO: add function that handles sync between cameras, osc, pulses, with env flag to determine if it should be used # SH: utils.py->class TriggerSynchronisation
        # TODO: add function that dumps all collected data to a single file
        # TODO: implement a multiple shot function that is called in get observables
        # TODO: add elegant way of working with two cameras + type checking (can't have two of the same camera type, etc.) # SH: screen.py->check_same_camera_type(screen1,screen2)
        # TODO: specify zero operating phase attribute for reference
        # TODO: add a check for zero phase after LLRF1 amplitude is changed
        # 
        
        results = {}
        metadata = kwargs or {}

        # you could save it to the same h5    
        if isinstance(self.image_diagnostic2, AWAImageDiagnostic):
            result_vir_cath = {}
            result_vir_cath.update(self.image_diagnostic2.measure_beamsize(
                1,False,**metadata
            ))
            results["VirtCath"] = result_vir_cath            

        # lists do not work here
        results["ICT1"] = caget("AWAVXI11ICT:Ch1")

        results["ICT1_nC"] = caget("AWAVXI11ICT:Ch1") * 1e9

        
        #results["FCT1"] = caget("AWAVXI11ICT:Ch3")

        
        #results["FCT1_nC"] = caget("AWAVXI11ICT:Ch3") * 1e9
            
              

        ### saving waveform to metadata since Xopt is not happy with lists of different lengths
        if "wfs" in observable_names:
            metadata["ICT1wf"] = np.array(caget("AWAVXI11ICT:rawwf1"))
            metadata["gun_waveform"] = np.array(caget("AWALLRF:WF:K1"))
            #metadata["FCT1wf"] = np.array(caget("AWAVXI11ICT:rawwf3"))
            metadata["FCT1wf_12.5GHz"] = np.array(caget("AWAVXI11:Tek1:wf1"))
            metadata["gunpulsewf"] = np.array(caget("AWABPPM:wf1"))
        

        #metadata["dyag_raw"] = np.array(caget("AWANIFG:ImgData"))
        #metadata["vcat_raw"] = np.array(caget("13ARV1:image1:ArrayData"))
        
        # check if observable names has a screen observable in it
        if len([ele for ele in observable_names if ele in screen_observables]):
            self.set_screen()
            
            #metadata = self.setpoints or {}
            metadata = metadata | results
    
            
            results.update(self.image_diagnostic.measure_beamsize(
                self.n_shots,False, **metadata
            ))
            
        
        
        results["time"] = time.time()

        #del metadata["dyag_raw"]
        #del metadata["vcat_raw"]
        
        #print(results)
        return results

    def get_variables(self, variable_names: List[str]) -> Dict:
        results = caget_many([pvs[name] for name in variable_names])
        return dict(zip(variable_names, results))

    def set_variables(self, variable_inputs: Dict[str, float]):
        if self.setpoints is not None:
            self.setpoints.update(variable_inputs)
        else:
            self.setpoints = variable_inputs
        for name, val in variable_inputs.items():
            # remove this
            if name ==  'zeroGunPhase': 
                continue
            caput(pvs[name], val)

    # TODO: fold into sychronize code
    def check_trigger():
        pv_ICT =pv.PV("AWAVXI11ICT:Ch1")
    
        temp = pv_ICT.get()
        ts_ict_ts_b =pv_ICT.timestamp
            # trigger scopes and camera update
            #not triggering pv_Trigger.put(1)
        time.sleep(1.1)
        ts_ict_ts_a = pv_ICT.timestamp
    
        if ((ts_ict_ts_a - ts_ict_ts_b) >  0.000001):
            print("Please, Make sure the osciloscopes and the Images are frozen after each trigger.")
            input("Press Enter to continue...")
            check_trigger()
        pass
